@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2007, 2018, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2016, 2019, MariaDB Corporation.
+Copyright (c) 2016, 2020, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -35,9 +35,10 @@ Completed 2011/7/10 Sunny and Jimmy Yang
 #include "srv0start.h"
 #include "ut0list.h"
 #include "zlib.h"
+#include "fts0opt.h"
 
 /** The FTS optimize thread's work queue. */
-static ib_wqueue_t* fts_optimize_wq;
+ib_wqueue_t* fts_optimize_wq;
 
 /** The FTS vector to store fts_slot_t */
 static ib_vector_t*  fts_slots;
@@ -168,8 +169,8 @@ struct fts_encode_t {
 /** We use this information to determine when to start the optimize
 cycle for a table. */
 struct fts_slot_t {
-	/** table identifier, or 0 if the slot is empty */
-	table_id_t	table_id;
+	/** table, or NULL if the slot is unused */
+	dict_table_t*	table;
 
 	/** whether this slot is being processed */
 	bool		running;
@@ -524,7 +525,7 @@ fts_index_fetch_nodes(
 	for (;;) {
 		error = fts_eval_sql(trx, *graph);
 
-		if (error == DB_SUCCESS) {
+		if (UNIV_LIKELY(error == DB_SUCCESS)) {
 			fts_sql_commit(trx);
 
 			break;				/* Exit the loop. */
@@ -537,7 +538,7 @@ fts_index_fetch_nodes(
 
 				trx->error_state = DB_SUCCESS;
 			} else {
-				ib::error() << "(" << ut_strerr(error)
+				ib::error() << "(" << error
 					<< ") while reading FTS index.";
 
 				break;			/* Exit the loop. */
@@ -859,7 +860,7 @@ fts_index_fetch_words(
 				error = fts_eval_sql(optim->trx, graph);
 			}
 
-			if (error == DB_SUCCESS) {
+			if (UNIV_LIKELY(error == DB_SUCCESS)) {
 				//FIXME fts_sql_commit(optim->trx);
 				break;
 			} else {
@@ -876,7 +877,7 @@ fts_index_fetch_words(
 
 					optim->trx->error_state = DB_SUCCESS;
 				} else {
-					ib::error() << "(" << ut_strerr(error)
+					ib::error() << "(" << error
 						<< ") while reading document.";
 
 					break;	/* Exit the loop. */
@@ -919,7 +920,7 @@ fts_fetch_doc_ids(
 	int		i = 0;
 	sel_node_t*	sel_node = static_cast<sel_node_t*>(row);
 	fts_doc_ids_t*	fts_doc_ids = static_cast<fts_doc_ids_t*>(user_arg);
-	fts_update_t*	update = static_cast<fts_update_t*>(
+	doc_id_t*	update = static_cast<doc_id_t*>(
 		ib_vector_push(fts_doc_ids->doc_ids, NULL));
 
 	for (exp = sel_node->select_list;
@@ -935,8 +936,7 @@ fts_fetch_doc_ids(
 		/* Note: The column numbers below must match the SELECT. */
 		switch (i) {
 		case 0: /* DOC_ID */
-			update->fts_indexes = NULL;
-			update->doc_id = fts_read_doc_id(
+			*update = fts_read_doc_id(
 				static_cast<byte*>(data));
 			break;
 
@@ -999,12 +999,12 @@ fts_table_fetch_doc_ids(
 	error = fts_eval_sql(trx, graph);
 	fts_sql_commit(trx);
 
-	mutex_enter(&dict_sys->mutex);
+	mutex_enter(&dict_sys.mutex);
 	que_graph_free(graph);
-	mutex_exit(&dict_sys->mutex);
+	mutex_exit(&dict_sys.mutex);
 
 	if (error == DB_SUCCESS) {
-		ib_vector_sort(doc_ids->doc_ids, fts_update_doc_id_cmp);
+		ib_vector_sort(doc_ids->doc_ids, fts_doc_id_cmp);
 	}
 
 	if (alloc_bk_trx) {
@@ -1021,7 +1021,7 @@ Do a binary search for a doc id in the array
 int
 fts_bsearch(
 /*========*/
-	fts_update_t*	array,	/*!< in: array to sort */
+	doc_id_t*	array,	/*!< in: array to sort */
 	int		lower,	/*!< in: the array lower bound */
 	int		upper,	/*!< in: the array upper bound */
 	doc_id_t	doc_id)	/*!< in: the doc id to search for */
@@ -1035,9 +1035,9 @@ fts_bsearch(
 		while (lower < upper) {
 			int	i = (lower + upper) >> 1;
 
-			if (doc_id > array[i].doc_id) {
+			if (doc_id > array[i]) {
 				lower = i + 1;
-			} else if (doc_id < array[i].doc_id) {
+			} else if (doc_id < array[i]) {
 				upper = i - 1;
 			} else {
 				return(i); /* Found. */
@@ -1046,7 +1046,7 @@ fts_bsearch(
 	}
 
 	if (lower == upper && lower < orig_size) {
-		if (doc_id == array[lower].doc_id) {
+		if (doc_id == array[lower]) {
 			return(lower);
 		} else if (lower == 0) {
 			return(-1);
@@ -1073,7 +1073,7 @@ fts_optimize_lookup(
 {
 	int		pos;
 	int		upper = static_cast<int>(ib_vector_size(doc_ids));
-	fts_update_t*	array = (fts_update_t*) doc_ids->data;
+	doc_id_t*	array = (doc_id_t*) doc_ids->data;
 
 	pos = fts_bsearch(array, static_cast<int>(lower), upper, first_doc_id);
 
@@ -1086,10 +1086,10 @@ fts_optimize_lookup(
 		/* If i is 1, it could be first_doc_id is less than
 		either the first or second array item, do a
 		double check */
-		if (i == 1 && array[0].doc_id <= last_doc_id
-		    && first_doc_id < array[0].doc_id) {
+		if (i == 1 && array[0] <= last_doc_id
+		    && first_doc_id < array[0]) {
 			pos = 0;
-		} else if (i < upper && array[i].doc_id <= last_doc_id) {
+		} else if (i < upper && array[i] <= last_doc_id) {
 
 			/* Check if the "next" doc id is within the
 			first & last doc id of the node. */
@@ -1228,12 +1228,12 @@ test_again:
 		delta for decoding the entries following this document's
 		entries. */
 		if (*del_pos >= 0 && *del_pos < (int) ib_vector_size(del_vec)) {
-			fts_update_t*	update;
+			doc_id_t*	update;
 
-			update = (fts_update_t*) ib_vector_get(
+			update = (doc_id_t*) ib_vector_get(
 				del_vec, ulint(*del_pos));
 
-			del_doc_id = update->doc_id;
+			del_doc_id = *update;
 		}
 
 		if (enc->src_ilist_ptr == src_node->ilist && doc_id == 0) {
@@ -1356,12 +1356,6 @@ fts_optimize_word(
 	enc.src_last_doc_id = 0;
 	enc.src_ilist_ptr = NULL;
 
-	if (fts_enable_diag_print) {
-		word->text.f_str[word->text.f_len] = 0;
-		ib::info() << "FTS_OPTIMIZE: optimize \"" << word->text.f_str
-			<< "\"";
-	}
-
 	while (i < size) {
 		ulint		copied;
 		fts_node_t*	src_node;
@@ -1439,11 +1433,6 @@ fts_optimize_write_word(
 
 	ut_ad(fts_table->charset);
 
-	if (fts_enable_diag_print) {
-		ib::info() << "FTS_OPTIMIZE: processed \"" << word->f_str
-			<< "\"";
-	}
-
 	pars_info_bind_varchar_literal(
 		info, "word", word->f_str, word->f_len);
 
@@ -1461,8 +1450,8 @@ fts_optimize_write_word(
 
 	error = fts_eval_sql(trx, graph);
 
-	if (error != DB_SUCCESS) {
-		ib::error() << "(" << ut_strerr(error) << ") during optimize,"
+	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
+		ib::error() << "(" << error << ") during optimize,"
 			" when deleting a word from the FTS index.";
 	}
 
@@ -1485,8 +1474,8 @@ fts_optimize_write_word(
 			error = fts_write_node(
 				trx, &graph, fts_table, word, node);
 
-			if (error != DB_SUCCESS) {
-				ib::error() << "(" << ut_strerr(error) << ")"
+			if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
+				ib::error() << "(" << error << ")"
 					" during optimize, while adding a"
 					" word to the FTS index.";
 			}
@@ -1875,9 +1864,8 @@ fts_optimize_index_completed(
 	error = fts_config_set_index_value(
 		optim->trx, index, FTS_LAST_OPTIMIZED_WORD, &word);
 
-	if (error != DB_SUCCESS) {
-
-		ib::error() << "(" << ut_strerr(error) << ") while updating"
+	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
+		ib::error() << "(" << error << ") while updating"
 			" last optimized word!";
 	}
 
@@ -2019,7 +2007,7 @@ fts_optimize_purge_deleted_doc_ids(
 	ulint		i;
 	pars_info_t*	info;
 	que_t*		graph;
-	fts_update_t*	update;
+	doc_id_t*	update;
 	doc_id_t	write_doc_id;
 	dberr_t		error = DB_SUCCESS;
 	char		deleted[MAX_FULL_NAME_LEN];
@@ -2029,11 +2017,11 @@ fts_optimize_purge_deleted_doc_ids(
 
 	ut_a(ib_vector_size(optim->to_delete->doc_ids) > 0);
 
-	update = static_cast<fts_update_t*>(
+	update = static_cast<doc_id_t*>(
 		ib_vector_get(optim->to_delete->doc_ids, 0));
 
 	/* Convert to "storage" byte order. */
-	fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
+	fts_write_doc_id((byte*) &write_doc_id, *update);
 
 	/* This is required for the SQL parser to work. It must be able
 	to find the following variables. So we do it twice. */
@@ -2055,11 +2043,11 @@ fts_optimize_purge_deleted_doc_ids(
 	/* Delete the doc ids that were copied at the start. */
 	for (i = 0; i < ib_vector_size(optim->to_delete->doc_ids); ++i) {
 
-		update = static_cast<fts_update_t*>(ib_vector_get(
+		update = static_cast<doc_id_t*>(ib_vector_get(
 			optim->to_delete->doc_ids, i));
 
 		/* Convert to "storage" byte order. */
-		fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
+		fts_write_doc_id((byte*) &write_doc_id, *update);
 
 		fts_bind_doc_id(info, "doc_id1", &write_doc_id);
 
@@ -2391,14 +2379,7 @@ fts_optimize_table_bk(
 		return(DB_SUCCESS);
 	}
 
-	dict_table_t* table = dict_table_open_on_id(
-		slot->table_id, FALSE, DICT_TABLE_OP_NORMAL);
-
-	if (!table) {
-		slot->last_run = now;
-		return DB_SUCCESS;
-	}
-
+	dict_table_t*	table = slot->table;
 	dberr_t		error;
 
 	if (fil_table_accessible(table)
@@ -2418,8 +2399,6 @@ fts_optimize_table_bk(
 		error = DB_SUCCESS;
 	}
 
-	dict_table_close(table, FALSE, FALSE);
-
 	return(error);
 }
 /*********************************************************************//**
@@ -2438,7 +2417,7 @@ fts_optimize_table(
 	fts_optimize_t*	optim = NULL;
 	fts_t*		fts = table->fts;
 
-	if (fts_enable_diag_print) {
+	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
 		ib::info() << "FTS start optimize " << table->name;
 	}
 
@@ -2490,7 +2469,7 @@ fts_optimize_table(
 		if (error == DB_SUCCESS
 		    && optim->n_completed == ib_vector_size(fts->indexes)) {
 
-			if (fts_enable_diag_print) {
+			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
 				ib::info() << "FTS_OPTIMIZE: Completed"
 					" Optimize, cleanup DELETED table";
 			}
@@ -2513,7 +2492,7 @@ fts_optimize_table(
 
 	fts_optimize_free(optim);
 
-	if (fts_enable_diag_print) {
+	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
 		ib::info() << "FTS end optimize " << table->name;
 	}
 
@@ -2553,12 +2532,24 @@ void fts_optimize_add_table(dict_table_t* table)
 		return;
 	}
 
+	/* If there is no fts index present then don't add to
+	optimize queue. */
+	if (!ib_vector_size(table->fts->indexes)) {
+		return;
+	}
+
 	/* Make sure table with FTS index cannot be evicted */
 	dict_table_prevent_eviction(table);
 
 	msg = fts_optimize_create_msg(FTS_MSG_ADD_TABLE, table);
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap);
+	mutex_enter(&fts_optimize_wq->mutex);
+
+	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+
+	table->fts->in_queue = true;
+
+	mutex_exit(&fts_optimize_wq->mutex);
 }
 
 /**********************************************************************//**
@@ -2585,6 +2576,13 @@ fts_optimize_remove_table(
 		return;
 	}
 
+	mutex_enter(&fts_optimize_wq->mutex);
+
+	if (!table->fts->in_queue) {
+		mutex_exit(&fts_optimize_wq->mutex);
+		return;
+	}
+
 	msg = fts_optimize_create_msg(FTS_MSG_DEL_TABLE, NULL);
 
 	/* We will wait on this event until signalled by the consumer. */
@@ -2597,11 +2595,17 @@ fts_optimize_remove_table(
 	remove->event = event;
 	msg->ptr = remove;
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap);
+	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+
+	mutex_exit(&fts_optimize_wq->mutex);
 
 	os_event_wait(event);
 
 	os_event_destroy(event);
+
+	ut_d(mutex_enter(&fts_optimize_wq->mutex));
+	ut_ad(!table->fts->in_queue);
+	ut_d(mutex_exit(&fts_optimize_wq->mutex));
 }
 
 /** Send sync fts cache for the table.
@@ -2610,9 +2614,6 @@ void
 fts_optimize_request_sync_table(
 	dict_table_t*	table)
 {
-	fts_msg_t*	msg;
-	table_id_t*	table_id;
-
 	/* if the optimize system not yet initialized, return */
 	if (!fts_optimize_wq) {
 		return;
@@ -2625,35 +2626,34 @@ fts_optimize_request_sync_table(
 		return;
 	}
 
-	msg = fts_optimize_create_msg(FTS_MSG_SYNC_TABLE, NULL);
+	fts_msg_t* msg = fts_optimize_create_msg(FTS_MSG_SYNC_TABLE, table);
 
-	table_id = static_cast<table_id_t*>(
-		mem_heap_alloc(msg->heap, sizeof(table_id_t)));
-	*table_id = table->id;
-	msg->ptr = table_id;
+	mutex_enter(&fts_optimize_wq->mutex);
 
-	ib_wqueue_add(fts_optimize_wq, msg, msg->heap);
+	ib_wqueue_add(fts_optimize_wq, msg, msg->heap, true);
+
+	mutex_exit(&fts_optimize_wq->mutex);
 }
 
 /** Add a table to fts_slots if it doesn't already exist. */
 static bool fts_optimize_new_table(dict_table_t* table)
 {
+	ut_ad(table);
+
 	ulint		i;
 	fts_slot_t*	slot;
 	fts_slot_t*	empty = NULL;
-	const table_id_t table_id = table->id;
-	ut_ad(table_id);
 
 	/* Search for duplicates, also find a free slot if one exists. */
 	for (i = 0; i < ib_vector_size(fts_slots); ++i) {
 
 		slot = static_cast<fts_slot_t*>(ib_vector_get(fts_slots, i));
 
-		if (!slot->table_id) {
+		if (!slot->table) {
 			empty = slot;
-		} else if (slot->table_id == table_id) {
+		} else if (slot->table == table) {
 			/* Already exists in our optimize queue. */
-			return(FALSE);
+			return false;
 		}
 	}
 
@@ -2662,36 +2662,35 @@ static bool fts_optimize_new_table(dict_table_t* table)
 
 	memset(slot, 0x0, sizeof(*slot));
 
-	slot->table_id = table->id;
-	slot->running = false;
-
-	return(TRUE);
+	slot->table = table;
+	return true;
 }
 
 /** Remove a table from fts_slots if it exists.
 @param[in,out]	table	table to be removed from fts_slots */
 static bool fts_optimize_del_table(const dict_table_t* table)
 {
-	const table_id_t table_id = table->id;
-	ut_ad(table_id);
-
+	ut_ad(table);
 	for (ulint i = 0; i < ib_vector_size(fts_slots); ++i) {
 		fts_slot_t*	slot;
 
 		slot = static_cast<fts_slot_t*>(ib_vector_get(fts_slots, i));
 
-		if (slot->table_id == table_id) {
-			if (fts_enable_diag_print) {
+		if (slot->table == table) {
+			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
 				ib::info() << "FTS Optimize Removing table "
 					<< table->name;
 			}
 
-			slot->table_id = 0;
-			return(TRUE);
+			mutex_enter(&fts_optimize_wq->mutex);
+			slot->table->fts->in_queue = false;
+			mutex_exit(&fts_optimize_wq->mutex);
+			slot->table = NULL;
+			return true;
 		}
 	}
 
-	return(FALSE);
+	return false;
 }
 
 /**********************************************************************//**
@@ -2705,7 +2704,7 @@ static ulint fts_optimize_how_many()
 	for (ulint i = 0; i < ib_vector_size(fts_slots); ++i) {
 		const fts_slot_t* slot = static_cast<const fts_slot_t*>(
 			ib_vector_get_const(fts_slots, i));
-		if (slot->table_id == 0) {
+		if (!slot->table) {
 			continue;
 		}
 
@@ -2741,22 +2740,14 @@ static bool fts_is_sync_needed()
 		const fts_slot_t* slot = static_cast<const fts_slot_t*>(
 			ib_vector_get_const(fts_slots, i));
 
-		if (slot->table_id == 0) {
+		if (!slot->table) {
 			continue;
 		}
 
-		dict_table_t* table = dict_table_open_on_id(
-			slot->table_id, FALSE, DICT_TABLE_OP_NORMAL);
-		if (!table) {
-			continue;
+		if (slot->table->fts && slot->table->fts->cache) {
+			total_memory += slot->table->fts->cache->total_size;
 		}
 
-		if (table->fts && table->fts->cache) {
-			total_memory += table->fts->cache->total_size;
-		}
-
-		dict_table_close(table, FALSE, FALSE);
-
 		if (total_memory > fts_max_total_cache_size) {
 			return(true);
 		}
@@ -2766,18 +2757,14 @@ static bool fts_is_sync_needed()
 }
 
 /** Sync fts cache of a table
-@param[in]	table_id	table id */
-static void fts_optimize_sync_table(table_id_t table_id)
+@param[in,out]	table	table to be synced */
+static void fts_optimize_sync_table(dict_table_t* table)
 {
-	if (dict_table_t* table = dict_table_open_on_id(
-		    table_id, FALSE, DICT_TABLE_OP_NORMAL)) {
-		if (fil_table_accessible(table)
-		    && table->fts && table->fts->cache) {
-			fts_sync_table(table, false);
-		}
-
-		dict_table_close(table, FALSE, FALSE);
+	if (table->fts && table->fts->cache && fil_table_accessible(table)) {
+		fts_sync_table(table, false);
 	}
+
+	DBUG_EXECUTE_IF("ib_optimize_wq_hang", os_thread_sleep(6000000););
 }
 
 /**********************************************************************//**
@@ -2803,8 +2790,7 @@ DECLARE_THREAD(fts_optimize_thread)(
 	/* Assign number of tables added in fts_slots_t to n_tables */
 	n_tables = ib_vector_size(fts_slots);
 
-	while (!done && srv_shutdown_state == SRV_SHUTDOWN_NONE) {
-
+	while (!done && srv_shutdown_state <= SRV_SHUTDOWN_INITIATED) {
 		/* If there is no message in the queue and we have tables
 		to optimize then optimize the tables. */
 
@@ -2816,7 +2802,7 @@ DECLARE_THREAD(fts_optimize_thread)(
 				ib_vector_get(fts_slots, current));
 
 			/* Handle the case of empty slots. */
-			if (slot->table_id) {
+			if (slot->table) {
 				slot->running = true;
 				fts_optimize_table_bk(slot);
 			}
@@ -2875,7 +2861,7 @@ DECLARE_THREAD(fts_optimize_thread)(
 					os_thread_sleep(300000););
 
 				fts_optimize_sync_table(
-					*static_cast<table_id_t*>(msg->ptr));
+					static_cast<dict_table_t*>(msg->ptr));
 				break;
 
 			default:
@@ -2894,8 +2880,8 @@ DECLARE_THREAD(fts_optimize_thread)(
 			fts_slot_t* slot = static_cast<fts_slot_t*>(
 				ib_vector_get(fts_slots, i));
 
-			if (table_id_t table_id = slot->table_id) {
-				fts_optimize_sync_table(table_id);
+			if (slot->table) {
+				fts_optimize_sync_table(slot->table);
 			}
 		}
 	}
@@ -2923,7 +2909,6 @@ fts_optimize_init(void)
 {
 	mem_heap_t*	heap;
 	ib_alloc_t*     heap_alloc;
-	dict_table_t*   table;
 
 	ut_ad(!srv_read_only_mode);
 
@@ -2939,31 +2924,25 @@ fts_optimize_init(void)
 	heap_alloc = ib_heap_allocator_create(heap);
 	fts_slots = ib_vector_create(heap_alloc, sizeof(fts_slot_t), 4);
 
-	/* Add fts tables to the fts_slots vector which were skipped during restart */
-	std::vector<dict_table_t*> table_vector;
-	std::vector<dict_table_t*>::iterator it;
-
-	mutex_enter(&dict_sys->mutex);
-	for (table = UT_LIST_GET_FIRST(dict_sys->table_LRU);
-             table != NULL;
-             table = UT_LIST_GET_NEXT(table_LRU, table)) {
-                if (table->fts &&
-                    dict_table_has_fts_index(table)) {
-			if (fts_optimize_new_table(table)){
-				table_vector.push_back(table);
-			}
+	/* Add fts tables to fts_slots which could be skipped
+	during dict_load_table_one() because fts_optimize_thread
+	wasn't even started. */
+	mutex_enter(&dict_sys.mutex);
+	for (dict_table_t* table = UT_LIST_GET_FIRST(dict_sys.table_LRU);
+	     table != NULL;
+	     table = UT_LIST_GET_NEXT(table_LRU, table)) {
+		if (!table->fts || !dict_table_has_fts_index(table)) {
+			continue;
 		}
-	}
 
-	/* It is better to call dict_table_prevent_eviction()
-	outside the above loop because it operates on
-	dict_sys->table_LRU list.*/
-	for (it=table_vector.begin();it!=table_vector.end();++it) {
-		dict_table_prevent_eviction(*it);
+		/* fts_optimize_thread is not started yet. So there is no
+		need to acquire fts_optimize_wq->mutex for adding the fts
+		table to the fts slots. */
+		ut_ad(!table->can_be_evicted);
+		fts_optimize_new_table(table);
+		table->fts->in_queue = true;
 	}
-
-	mutex_exit(&dict_sys->mutex);
-	table_vector.clear();
+	mutex_exit(&dict_sys.mutex);
 
 	fts_opt_shutdown_event = os_event_create(0);
 	last_check_sync_time = time(NULL);
