@@ -54,10 +54,6 @@ savepoint. */
 @return	old mode */
 #define mtr_set_log_mode(m, d)	(m)->set_log_mode((d))
 
-/** Read 1 - 4 bytes from a file page buffered in the buffer pool.
-@return	value read */
-#define mtr_read_ulint(p, t, m)	(m)->read_ulint((p), (t))
-
 /** Release an object in the memo stack.
 @return true if released */
 #define mtr_memo_release(m, o, t)					\
@@ -239,13 +235,6 @@ struct mtr_t {
 	bool is_named_space(const fil_space_t* space) const;
 #endif /* UNIV_DEBUG */
 
-	/** Read 1 - 4 bytes from a file page buffered in the buffer pool.
-	@param ptr	pointer from where to read
-	@param type)	MLOG_1BYTE, MLOG_2BYTES, MLOG_4BYTES
-	@return	value read */
-	inline ulint read_ulint(const byte* ptr, mlog_id_t type) const
-		MY_ATTRIBUTE((warn_unused_result));
-
 	/** Acquire a tablespace X-latch.
 	@param[in]	space_id	tablespace ID
 	@param[in]	file		file name from where called
@@ -437,10 +426,6 @@ struct mtr_t {
 	static inline bool is_block_dirtied(const buf_block_t* block)
 		MY_ATTRIBUTE((warn_unused_result));
 
-	/** Get the buffer fix count for the block added by this mtr.
-	@param[in]	block	block to be checked
-	@return number of buffer count added by this mtr */
-	int32_t get_fix_count(const buf_block_t *block);
 private:
 	/** Prepare to write the mini-transaction log to the redo log buffer.
 	@return number of bytes to write in finish_write() */
